#!/usr/bin/perl -w
#
# wbuser - Administer winbind user mappings.
#
# Copyright (c) 2005-2006 Occam's Razor. All rights reserved.
#
# See the LICENSE file distributed with this code for restrictions on its use
# and further distribution.
# Original distribution available at <http://www.occam.com/tools/>.
#
# $Id: wbuser.pl,v 1.1 2006/01/13 01:14:02 leonvs Exp $
#
# TODO
#	handle group mappings too?
#	allow removal by UID or SID, as well as username?
#

use strict;
use Getopt::Std;
use vars qw($opt_h $opt_V $opt_d $opt_a $opt_r $opt_l);

# global variables
my $progname;
my $DEBUG = 0;
my $sambaDir = "/opt/local/samba";
my $wbinfoCmd = "$sambaDir/bin/wbinfo";
my $tdbtoolCmd = "$sambaDir/bin/tdbtool";
my $tdbdumpCmd = "$sambaDir/bin/tdbdump";
my $tdbbackupCmd = "$sambaDir/bin/tdbbackup";
my $idmapFile = "$sambaDir/var/locks/winbindd_idmap.tdb";
my $mode;
my $uname;
my $uid;

processArgs();
for ($mode) {
	/^LIST$/	and do { list(); last; };
	/^ADD$/		and do { add($uname, $uid); last; };
	/^REMOVE$/	and do { remove($uname); last; };
	report("ERR", "Unknown mode: $mode.");
}

sub processArgs {
	$progname = $0; $progname =~ s/.*\///;

	my $VERSION = "$progname 1.1\n";

	my $USAGE = <<EOF;
$VERSION
usage:	$progname { -h | -V }
	$progname [-d] -l
	$progname [-d] -a username UID
	$progname [-d] -r username

	-h: Prints this usage statement and exits.
	-V: Prints version number and exits.
	-d: Prints debugging output.
	-l: Lists current winbind user mappings.
	-a: Adds a winbind mapping for the specified Windows username to the
		specified UNIX UID.
	-r: Removes the winbind mapping for the specified user.
EOF

	getopts('hVdlar') or die $USAGE;
	$opt_h and die $USAGE;
	$opt_V and die $VERSION;

	$opt_d and ++$DEBUG;

	# Only one of the following options should be set.
	$opt_l xor $opt_a xor $opt_r or die $USAGE;
	if ($opt_l) {
		$mode = "LIST";
	} elsif ($opt_a) {
		$uname = shift @ARGV or die $USAGE;
		isValidUsername($uname) or die $USAGE;
		$uid = shift @ARGV or die $USAGE;
		# Allow for 64-bit UIDs.
		$uid =~ /^\d{1,20}$/ or die $USAGE;
		$mode = "ADD";
	} elsif ($opt_r) {
		$uname = shift @ARGV or die $USAGE;
		isValidUsername($uname) or die $USAGE;
		$mode = "REMOVE";
	}

	# Should be no more arguments left.
	@ARGV and die $USAGE;
}

sub list {
	my $uid;
	my %uid = ();
	my %sid = ();

	open TDB, "$tdbdumpCmd $idmapFile |" or
		report("ERR", "Couldn't dump contents of $idmapFile.");
	while (<TDB>) {
		# Look for key first, then look for data.
		if ($uid) {
			/^data/ or next;
			# Data should be of the form:
			# data = "S-[mix of digits & dashes]\00"
			/^data = \"(S\-[\d\-]+)\\/ and my $sid = $+;

			my $uname = `$wbinfoCmd -s $sid`;
			chomp $uname;
			# Keep only first term in output.
			$uname =~ s/\s.*$//;

			$uid{$uname} = $uid;
			$sid{$uname} = $sid;
			undef $uid;
		} else {
			/^key/ or next;
			# Only look for keys of the form:
			# key = "UID NNNNN\00"
			/^key = \"UID (\d+)\\/ and $uid = $+;
		}
	}
	close TDB;

	print "Username\tUID\tSID\n";
	foreach my $uname (sort keys %uid) {
		print "$uname\t$uid{$uname}\t$sid{$uname}\n";
	}
}

sub add {
	my $uname = shift;
	my $uid = shift;

	my $sid = `$wbinfoCmd -n $uname`;
	chomp $sid;
	$sid =~ /^S\-[\d\-]+\s/ or
		report("ERR", "Lookup failed for $uname: $sid");
	# Keep only first term in output.
	$sid =~ s/\s.*$//;

	system "$tdbbackupCmd $idmapFile";

	my $uidIn = <<EOF;
open $idmapFile
show UID\\ $uid
quit
EOF
	my $uidOut = `echo "$uidIn" | $tdbtoolCmd`;
	report("DEBUG", $uidOut);
	$uidOut =~ /fetch failed/ or
		report("ERR", "Entry already exists for UID $uid.");

	my $sidIn = <<EOF;
open $idmapFile
show $sid
quit
EOF
	my $sidOut = `echo "$sidIn" | $tdbtoolCmd`;
	report("DEBUG", $sidOut);
	$sidOut =~ /fetch failed/ or
		report("ERR", "Entry already exists for SID $sid.");

	my $addIn = <<EOF;
open $idmapFile
insert UID\\ $uid $sid
insert $sid UID\\ $uid
quit
EOF
	my $addOut = `echo "$addIn" | $tdbtoolCmd`;
	report("DEBUG", $addOut);

	# Check to see if entries were added.
	$uidOut = `echo "$uidIn" | $tdbtoolCmd`;
	report("DEBUG", $uidOut);
	$uidOut =~ /fetch failed/ and
		report("ERR", "Add failed for UID $uid.");
	$sidOut = `echo "$sidIn" | $tdbtoolCmd`;
	report("DEBUG", $sidOut);
	$sidOut =~ /fetch failed/ and
		report("ERR", "Add failed for SID $sid.");

	report("INFO", "Add successful for $uname, UID $uid, SID $sid.");
}

sub remove {
	my $uname = shift;

	my $sid = `$wbinfoCmd -n $uname`;
	chomp $sid;
	$sid =~ /^S\-[\d\-]+\s/ or
		report("ERR", "Lookup failed for $uname: $sid");
	# Keep only first term in output.
	$sid =~ s/\s.*$//;

	my $uid = `$wbinfoCmd -S $sid`;
	chomp $uid;
	$uid =~ /^\d+$/ or
		report("ERR", "Lookup failed for $sid: $uid");

	system "$tdbbackupCmd $idmapFile";

	my $removeIn = <<EOF;
open $idmapFile
delete UID\\ $uid
delete $sid
quit
EOF
	my $removeOut = `echo "$removeIn" | $tdbtoolCmd`;
	report("DEBUG", $removeOut);

	my $uidIn = <<EOF;
open $idmapFile
show UID\\ $uid
quit
EOF
	my $uidOut = `echo "$uidIn" | $tdbtoolCmd`;
	report("DEBUG", $uidOut);
	$uidOut =~ /fetch failed/ or
		report("ERR", "Entry still exists for UID $uid.");

	my $sidIn = <<EOF;
open $idmapFile
show $sid
quit
EOF
	my $sidOut = `echo "$sidIn" | $tdbtoolCmd`;
	report("DEBUG", $sidOut);
	$sidOut =~ /fetch failed/ or
		report("ERR", "Entry still exists for SID $sid.");

	report("INFO", "Removal successful for $uname, UID $uid, SID $sid.");
}

sub isValidUsername {
	my $username = shift;

	# Username should include only alphanumerics (including underscores),
	# hyphens, plusses, and dots. This may be too restrictive, but better
	# safe than sorry.
	$username =~ /^[\w\.\-\+]+$/ or return 0;

	return 1;
}

sub report {
	my ($severity, $message) = @_;

	if	($severity eq "ERR")	{ die "$progname: $message\n"; }
	elsif	($severity eq "WARN")	{ warn "$progname: $message\n"; }
	elsif	($severity eq "INFO")	{ print "$progname: $message\n"; }
	elsif	($severity eq "DEBUG")	{ warn "$progname: DEBUG: $message\n"
						if $DEBUG; }
	else	{ die "$progname: Undefined error. $message\n"; }
}
